"""
Context: Toolbox || **Category: Fundamental**.

A category to group all of the fundamental indicators available in the
`Toolbox()`.

Fundamental indicators relies on earnings data, valuation models of companies,
balance sheet metrics etc...
"""
