"""
**Context: Toolbox**.

A category to group all of the technical indicators available in the `Toolbox()`

Technical indicators rely on statistical transformations of time series data.
These are raw math operations.
"""
