"""
The `core` module to contain logic & functions used in controllers.

This module is intended to contain sub-modules and functions that are not
directly utilized from the package, but rather used in building the package
itself. This means that the core module should not contain any code that is
specific to the package's use case, but rather should be generic and reusable
in other contexts.
"""
